/* SPDX-License-Identifier: GPL-2.0-or-later. */
#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif
#include <ctype.h>
#include <errno.h>
#include <error.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))

static const char* prefix_template = PREFIX "/lib/llvm-%s" "/bin";

static const struct assoc {
	const char *name;
	const char *value;
} table[] = {
	{ "c++", "clang++" },
	{ "cc", "clang" },
	{ "f77", "flang" },
	{ "f95", "flang" },
	{ "llvm-alt-tool-wrapper", (const char*)ENOENT /* print usage */ },
};

static int usage_and_die(void) {
	fprintf(stderr,
"You have invoked the ALT wrapper program for the installed\n"
"LLVM toolchains and utilities.\n"
"You're seeing this message, because it could not determine\n"
"which program it was executed as.\n"
"The package \"%s\" symlinks /usr/bin/clang and other tools\n"
"to this program, so it can find and execute the real tool.\n"
"\n"
"The wrapper can be controlled with environment variables.\n"
"To use a specific major version (e. g. 9.0) of a particular tool,\n"
"call its symlink with the environment variable ALTWRAP_LLVM_VERSION\n"
"set to \"9.0\". The default is \"%s\".\n"
"",
PACKAGE_NAME,
DEFAULT_VERSION);
	return EXIT_FAILURE;
}

static void logprintf(const char* fmt, ...) {
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "%s: ", program_invocation_short_name);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}

static void
friendly_perror(const char *s)
{
	int saved_errno = errno;
	if (saved_errno == ENOENT)
	{
		/* Advise the user. */
		logprintf("A symlink to the following path was installed along with the generic wrapper,\n"
			"but it looks like the actual tool is not installed. You might want\n"
			"to install the relevant package.\n");
	}

	errno = saved_errno;
	perror(s);
}

static char* leftchomp(const char* var) {
	if (var)
	{
		while (*var && isspace(*var))
			var++;
		if (*var)
			return (char*)var;
		return NULL;
	}
	return NULL;
}

/* If `name' exists in the environment, treat it as a PATH-like colon-separated
 * directory list and prepend `pre' to it.
 * If `name' does not exist in the environment, or is empty, or is
 * whitespace-only, set it to the value of `pre'.
 */
static void prepend_path_env(const char* name, const char* pre) {
	char* m;
	char* ss = leftchomp(getenv(name));
	if (ss)
	{
		if (asprintf(&m, "%s:%s", pre, ss) < 0)
			error(EXIT_FAILURE, errno, "asprintf");
	}
	else
	{
		if (asprintf(&m, "%s", pre) < 0)
			error(EXIT_FAILURE, errno, "asprintf");
	}
	setenv(name, m, 1);
}

int
main(int argc, char **argv)
{
	char   *real_executable = "", *llvm_bindir = "";
	char   *executable_name, *bin_name, *path;
	const char *call_name = program_invocation_short_name;
	const char *dry_run = getenv("ALTWRAP_DRY_RUN");
	/* Example: if ALTWRAP_LLVM_VERSION=11.0 and we were invoked as "clang",
	 * /usr/lib/llvm-11.0/bin/clang is getting called.
	 */
	const char *version = getenv("ALTWRAP_LLVM_VERSION");
	const char *use_ccache = getenv("ALTWRAP_LLVM_USE_CCACHE");
	size_t i;

	for (i = 0; i < ARRAY_SIZE(table); ++i)
		if (!strcmp(call_name, table[i].name))
		{
			call_name = table[i].value;
			break;
		}
	if (call_name == (const char*)ENOENT)
		return usage_and_die();

	version = leftchomp(version);
	if (!version) {
		version = DEFAULT_VERSION;
	}
	executable_name = strdup(call_name);

	argv[0] = executable_name;

	bin_name = executable_name;

	if (asprintf(&llvm_bindir, prefix_template, version) < 0)
		error(EXIT_FAILURE, errno, "asprintf");
	if (asprintf(&real_executable, "%s/%s", llvm_bindir, bin_name) < 0)
		error(EXIT_FAILURE, errno, "asprintf");

	use_ccache = leftchomp(use_ccache);
	if (use_ccache)
	{
		path = BINDIR "/ccache";
		prepend_path_env("CCACHE_PATH", llvm_bindir);
	}
	else
		path = real_executable;

	dry_run = leftchomp(dry_run);
	if (dry_run) {
		logprintf("path=%s\n", path);
		for (i = 0; i < argc; ++i) {
			logprintf("argv[%ld]=%s\n", i, argv[i]);
		}
		return EXIT_SUCCESS;
	}

	execv(path, argv);
	friendly_perror(path);
	return EXIT_FAILURE;
}
